﻿using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Newtonsoft.Json.Linq;
using OpenTK.Graphics.OpenGL;
using Vintagestory.API.Server;
using Vintagestory;
using System;
using Vintagestory.API.Config;

namespace nearsight
{
    /// <summary>
    /// Makes you all happy and giddy when you hold a chick in your hands
    /// A sample on how to load your own custom shader and how to render with it with a quad model during the ortho/2d gui render pass
    /// </summary>
    public class NearsightMod : ModSystem
    {
        ICoreClientAPI capi;
        IShaderProgram overlayShaderProg;
        NearsightednessRenderer renderer;

        public static NearsightednessSettings nearsightSettings;
        public static bool settingsLoaded = false;

        long gameTickListenerID;

        public override bool ShouldLoad(EnumAppSide side)
        {
            return side == EnumAppSide.Client;
        }

        public override void StartClientSide(ICoreClientAPI api)
        {
            this.capi = api;

            nearsightSettings = new NearsightednessSettings();

            api.Event.ReloadShader += LoadShader;
            LoadShader();

            renderer = new NearsightednessRenderer(api, overlayShaderProg);
            api.Event.RegisterRenderer(renderer, EnumRenderStage.AfterFinalComposition);

            //gameTickListenerID = api.Event.RegisterGameTickListener(TryRegister, 100);
        }

        public void TryRegister(float dt)
        {
            if (!settingsLoaded)
            {
                nearsightSettings = JsonUtil.FromBytes<NearsightednessSettings>(capi.World.Player.WorldData.GetModdata("nearsightSettings"));

                if (nearsightSettings == null)
                {
                    nearsightSettings = new NearsightednessSettings(0.95f, 20f, true);
                    capi.World.Player.WorldData.SetModdata("nearsightSettings", JsonUtil.ToBytes(nearsightSettings));
                }
                settingsLoaded = true;
                capi.SendChatMessage("Loaded nearsighted settings");
            }
            else
            {
                capi.Event.UnregisterGameTickListener(gameTickListenerID);
            }
        }

        public bool LoadShader()
        {
            overlayShaderProg = capi.Shader.NewShaderProgram();
            overlayShaderProg.VertexShader = capi.Shader.NewShader(EnumShaderType.VertexShader);
            overlayShaderProg.FragmentShader = capi.Shader.NewShader(EnumShaderType.FragmentShader);

            overlayShaderProg.VertexShader.Code = GetVertexShaderCode();
            overlayShaderProg.FragmentShader.Code = GetFragmentShaderCode();

            capi.Shader.RegisterMemoryShaderProgram("exampleoverlay", overlayShaderProg);
            //overlayShaderProg.PrepareUniformLocations("time");
            overlayShaderProg.Compile();

            if (renderer != null)
            {
                renderer.overlayShaderProg = overlayShaderProg;
            }

            return true;
        }


        #region Shader Code

        public string GetVertexShaderCode()
        {
            return @"
#version 330 core
#extension GL_ARB_explicit_attrib_location: enable

layout(location = 0) in vec3 vertex;

out vec2 uv;

void main(void)
{
    gl_Position = vec4(vertex.xy, 0, 1);
    uv = (vertex.xy + 1.0) / 2.0;
}
";
        }

        public string GetFragmentShaderCode()
        {
            return @"
#version 330 core

in vec2 uv;

out vec4 outColor;

uniform float time;
uniform float xs;
uniform float ys; // texture resolution
uniform float r; // radius
uniform float blurDistance;

uniform sampler2D depthTexture;
uniform sampler2D colorTexture;

float map(float value, float min1, float max1, float min2, float max2) {
  return min2 + (value - min1) * (max2 - min2) / (max1 - min1);
}

vec4 blur(sampler2D txr, vec2 pos)
{
    pos = vec2(pos[0] - 0.5, pos[1] - 0.5);

    float x,y,xx,yy,rr=r*r,dx,dy,w,w0;
    w0=0.3780/pow(r,1.975);
    vec2 p;
    vec4 col=vec4(0.0,0.0,0.0,0.0);
    for (dx=1.0/xs,x=-r,p.x=0.5+(pos.x)+(x*dx);x<=r;x++,p.x+=dx)
        {
        xx=x*x;
        for (dy=1.0/ys,y=-r,p.y=0.5+(pos.y)+(y*dy);y<=r;y++,p.y+=dy)
        {
            yy=y*y;
            if (xx+yy<=rr)
            {
                w=w0*exp((-xx-yy)/(2.0*rr));
                col+=texture2D(txr,p)*w;
            }
        }
    }
    return col;
}

void main () {
    vec4 depth = texture2D(depthTexture, uv);
    vec4 color = texture2D(colorTexture, uv);

    float col = depth[0];

    float lerpValue = clamp(map(depth[0], blurDistance, 1, 0, 1), 0, 1);

    vec4 blurred = blur(colorTexture, uv);

    outColor = vec4(blurred[0], blurred[1], blurred[2], lerpValue);
}
";
        }

        #endregion

    }

    [Serializable]
    public class NearsightednessSettings
    {
        public float MinBlurDistance;
        public float BlurRadius;
        public bool BlurEnabled;

        public NearsightednessSettings (float minBlurDistance, float blurRadius, bool blurEnabled)
        {
            MinBlurDistance = Math.Max(Math.Min(minBlurDistance, 1f), 0f);
            BlurRadius = Math.Max(blurRadius, 0);
            BlurEnabled = blurEnabled;
        }

        public NearsightednessSettings()
        {
            MinBlurDistance = 0;
            BlurRadius = 0;
            BlurEnabled = false;
        }

        public NearsightednessSettings(bool blurEnabled)
        {
            MinBlurDistance = 0.95f;
            BlurRadius = 20f;
            BlurEnabled = blurEnabled;
        }
    }


    public class NearsightednessRenderer : IRenderer
    {
        MeshRef quadRef;
        ICoreClientAPI capi;
        public IShaderProgram overlayShaderProg;


        public NearsightednessRenderer(ICoreClientAPI capi, IShaderProgram overlayShaderProg)
        {
            this.capi = capi;
            this.overlayShaderProg = overlayShaderProg;

            MeshData quadMesh = QuadMeshUtil.GetCustomQuadModelData(-1, -1, 0, 2, 2);
            quadMesh.Rgba = null;

            quadRef = capi.Render.UploadMesh(quadMesh);
        }

        public double RenderOrder
        {
            get { return 1.1; }
        }

        public int RenderRange { get { return 1; } }

        public void Dispose()
        {
            capi.Render.DeleteMesh(quadRef);
            overlayShaderProg.Dispose();
        }

        public void OnRenderFrame(float deltaTime, EnumRenderStage stage)
        {
            if (!NearsightMod.nearsightSettings.BlurEnabled) return;
            ItemStack stack = capi.World.Player.InventoryManager.GetInventory(capi.World.Player.InventoryManager.GetInventoryName(GlobalConstants.characterInvClassName))[0].Itemstack;
            if (stack != null && (stack.Collectible.Attributes?.IsTrue("isGlasses") ?? false)) return;


            IShaderProgram curShader = capi.Render.CurrentActiveShader;
            curShader?.Stop();

            overlayShaderProg.Use();

            capi.Render.GlToggleBlend(true);

            IClientPlayer player = capi.World.Player;
            //IServerPlayer splayer;

            overlayShaderProg.Uniform("time", capi.World.ElapsedMilliseconds / 1000f);
            overlayShaderProg.BindTexture2D("depthTexture", capi.Render.FrameBuffers[(int)EnumFrameBuffer.Primary].DepthTextureId, 0);

            overlayShaderProg.Uniform("xs", (float)capi.Render.FrameWidth);
            overlayShaderProg.Uniform("ys", (float)capi.Render.FrameHeight);
            overlayShaderProg.Uniform("r", NearsightMod.nearsightSettings.BlurRadius);
            overlayShaderProg.Uniform("blurDistance", NearsightMod.nearsightSettings.MinBlurDistance);
            overlayShaderProg.BindTexture2D("colorTexture", capi.Render.FrameBuffers[(int)EnumFrameBuffer.Primary].ColorTextureIds[0], 1);
            //overlayShaderProg.BindTexture2D("UITexture", capi.Render.FrameBuffers[(int)EnumFrameBuffer.Primary].ColorTextureIds[0], 0);

            // Primary : 0 : world, 1 : glow, 2 : Normal, 3 : eye space coords ?, 

            capi.Render.RenderMesh(quadRef);
            overlayShaderProg.Stop();


            curShader?.Use();
        }
    }
}